import json
import sys
import time
import re
import traceback

group_dic = dict()

with open('small.dic', 'r') as smd:
    mtch = r"<(\d+),\s(\d+)>"
    for line in smd:
        comps = [s.strip() for s in line.strip().split(';')]
        key = tuple(int(x) for x in re.match(mtch, comps[0]).groups())
        group_dic[key] = comps[1]


def f_permutation(plst, latex=False):
    imgs = [i-1 for i in plst]
    assert sum(imgs) == len(imgs)*(len(imgs)-1)/2
    visited = [False for _ in imgs]
    cycles = []
    for pt in range(len(imgs)):
        if visited[pt]:
            continue
        acycle = [pt]
        p = pt
        while acycle[0] != imgs[p]:
            p = imgs[p]
            acycle.append(p)
            visited[p] = True
        cycles.append(acycle)
    ret = ""
    for c in cycles:
        if len(c) > 1:
            ret += "({})".format(",".join([str(i+1) for i in c]))
    if ret == "":
        ret = "()"
    if latex:
        return '$' + ret + '$'
    return ret


def latexify_group(sdesc):
    replaced = sdesc
    replaced = re.sub(r"x", r"\\times", replaced)  # direct product
    replaced = re.sub(r":", r"\\colon", replaced)  # semidicrect product
    replaced = re.sub(r"C(\d+)", r"\\mathrm{C}_{\1}",
                      replaced)  # cyclic
    replaced = re.sub(r"(^D|\sD)(\d+)", r" \\mathrm{D}_{\2}",
                      replaced)  # dihedral
    replaced = re.sub(r"(^Q|\sQ)(\d+)", r" \\mathrm{Q}_{\2}",
                      replaced)  # quaternion
    replaced = re.sub(r"S(\d+)", r"\\mathrm{S}_{\1}",
                      replaced)  # symmetric
    replaced = re.sub(r"A(\d+)", r"\\mathrm{A}_{\1}",
                      replaced)  # symmetric

    replaced = re.sub(r"(\w+)\((.+)\)", r"\\mathrm{\1}(\2)",
                      replaced)  # other paren
    # replaced = re.sub(r"([A-Z]+)(.+)", r"\\mathrm{\1}_{\2}",
    #                   replaced)  # other index
    # print("DEBUG replaced", replaced)
    return '{{\\small ${}$}}'.format(replaced)


def f_group(o, c, latex=False, g_dic = group_dic):
    gstring = ""
    try:
        key = tuple([o, c])
        gstring = g_dic[key]
        if latex:
            gstring = latexify_group(gstring)
    except KeyError:
        pass
    if latex:
        return '$\\langle {}, {} \\rangle$'.format(o, c), gstring
    return '<{}, {}>'.format(o, c), gstring


def f_signature(g: int, bds: list, len_limit: int = 4, latex=False):
    bdstring = ""
    if len(bds) > len_limit:
        com = []
        bds.sort()
        for n in bds:
            found = False
            for i in range(len(com)):
                if com[i][0] == n:
                    com[i][1] = com[i][1]+1
                    found = True
                    break
            if not(found):
                com.append([n, 1])
        # print(com)
        fst = True
        for e in com:
            b = e[0]
            k = e[1]
            if not(fst):
                bdstring += ', '
            else:
                fst = False
            if k == 1:
                bdstring += str(b)
            else:
                bdstring += '{}^{}'.format(b, k)
    else:
        bdstring = ', '.join([str(i) for i in bds])

    rstring = "({}; -)".format(g) if bdstring == '' else "({}; {})".format(g, bdstring)
    if latex:
        return "${}$".format(rstring)
    return rstring

    # if latex:
    #     if bdstring == '':
    #         return "$({}; -)$".format(g, bdstring)
    #     else:
    #         return "$({}; {})$".format(g, bdstring)
    # return "({}; {})".format(g, bdstring)


def write_index(data, o_name, gamma=None, multi=None, latex=False,trivial=False):
    try:
        action_list = data['actions']
        genus = action_list[0]['genus']
    except Exception as e:
        raise e
    with open(o_name, 'w') as ostrm:
        if latex:
            uvtext = """%\\documentclass{article}

%\\usepackage{longtable}
%\\usepackage{amsmath,amsfonts,amssymb}


%\\begin{document}
\\begin{center}
\\begin{longtable}{|l|l|l|l|r|r|r|}"""
            uvtext += '\caption{{Census of finite group actions on Riemann surfaces of genus $g={}$}}\n\\label{{census:genus{}}} \\\\'.format(
                genus, genus)
            uvtext += """

\\hline \\multicolumn{1}{|c|}{\\textbf{Ref}} 
& \\multicolumn{1}{c|}{\\textbf{Signature}}
& \\multicolumn{1}{c|}{\\textbf{SMG id}}
& \\multicolumn{1}{c|}{\\textbf{Structure}}
& \\multicolumn{1}{c|}{\\textbf{epi}}
& \\multicolumn{1}{c|}{\\textbf{autg}}
& \\multicolumn{1}{c|}{$\\mathbf{\\sim}$\\textbf{top}}
 \\\\ \\hline 
\endfirsthead

\\multicolumn{7}{c}%
{{\\bfseries \\tablename\\ \\thetable{} -- continued from previous page}} \\\\
\\hline \\multicolumn{1}{|c|}{\\textbf{Ref}} 
& \\multicolumn{1}{c|}{\\textbf{Signature}}
& \\multicolumn{1}{c|}{\\textbf{SMG id}}
& \\multicolumn{1}{c|}{\\textbf{Structure}}
& \\multicolumn{1}{c|}{\\textbf{epi}}
& \\multicolumn{1}{c|}{\\textbf{autg}}
& \\multicolumn{1}{c|}{$\\mathbf{\\sim}$\\textbf{top}} \\\\ \\hline 
\\endhead

\\hline \\multicolumn{7}{|r|}{{Continued on next page}} \\\\ \\hline
\\endfoot

\\hline \\hline
\\endlastfoot
            """
        else:
            uvtext = "Census of actions of finite groups on Riemann surfaces of genus " \
                + str(genus) \
                + "\nExported on: " + time.asctime() \
                + "\nCreated on: " + data['created'] \
                + "\nCreator: rhsolver(Magma), ver: " + data['version'] \
                + "\n(c) 2023 Jan Karabas, Matej Bel University." \
                + "\nThe catalogue may be used only if you refer to the original source." \
                + "\nBibTeX entry:" \
                + "\n" + '-'*88 + "\n" \
                + r"@misc{jk23-1," + "\n" \
                + r"      author       = {Karab\'a\v s, J\'an}," + "\n" \
                + r"      title        = {Actions of finite groups on {R}iemann surfaces of higher genera}," + "\n" \
                + r"      year         = {2023}," + "\n" \
                + r"      howpublished = {\url{https://www.savbb.sk/~karabas/science/discactions.html}}" + "\n" \
                + r"}"  \
                + "\n" + '-'*88 + "\n" \
                + r"See also https://www.savbb.sk/~karabas/science/discactions.html for more information." \
                + "\n" + '='*88 + "\n"
        ostrm.write("{}\n\n".format(uvtext))
        ndx = 0
        if latex:
            pass
        else:

            ostrm.write("{}\n".format('-'*88))
            ostrm.write("{:7s} {:30s} {:10s} {:>12s} {:>12s} {:>12s}\n".format(
                'Code', 'Signature', 'SMG-ID', 'epi', 'autg', '~top'))
            ostrm.write("{}\n".format('-'*88))
        for elt in action_list:
            ndx += 1
            if not(gamma is None):
                if not(elt['gamma'] in gamma):
                    continue
            if 'top' in elt['enums']: 
                if not(multi is None):
                    if not(elt['enums']['top'] in multi):
                        continue
            # if gamma >= 0 and elt['gamma'] != gamma:
            #     continue
            # genus, gamma, bd, group, imgs, enums,
            s_num = 'O{}.{}'.format(elt['genus'], ndx)

            s_sig = f_signature(elt['gamma'], elt['bd'], latex=latex)
            s_gcode, s_grp = f_group(
                elt['group']['order'], elt['group']['code'], latex=latex)
            s_epi = '?' if not(
                'epi' in elt['enums']) else str(elt['enums']['epi'])
            s_autg = '?' if not(
                'autg' in elt['enums']) else str(elt['enums']['autg'])
            if trivial:
                if s_autg=='1':
                    if 'top' in elt['enums']:
                        if elt['enums']['top']!=1:
                            raise RuntimeError("autg != 1 and top == 1")
                    s_weak = '1'
                else:
                    s_weak = '?' if not('top' in elt['enums']) else str(elt['enums']['top'])
            else:            
                s_weak = '?' if not(
                    'top' in elt['enums']) else str(elt['enums']['top'])
            if latex:
                out_line = "{} & {} & {} & {} & ${}$ & ${}$ & ${}$ \\\\".format(
                    s_num, s_sig, s_gcode, s_grp, s_epi, s_autg, s_weak)
            else:
                out_line = "{:7s} {:30s} {:10s} {:>12s} {:>12s} {:>12s}".format(
                    s_num, s_sig, s_gcode, s_epi, s_autg, s_weak)
                if s_grp != '':
                    out_line += '    : {}'.format(s_grp)
            ostrm.write("{}\n".format(out_line))
        if latex:
            towrite = """\\end{longtable}
\\end{center}
%\\end{document}
"""
            ostrm.write(towrite)


def write_human_readable(data, o_name, gamma=None, multi=None,trivial=False):
    try:
        action_list = data['actions']
        genus = action_list[0]['genus']
    except Exception as e:
        raise e
    with open(o_name, 'w') as ostrm:
        uvtext = "Census of actions of finite groups on a surface of genus " \
            + str(genus) \
            + "\nExported on: " + time.asctime() \
            + "\nCreated on: " + data['created'] \
            + "\nCreator: rhsolver(Magma), ver: " + data['version'] \
            + "\n(c) 2023 Jan Karabas, Matej Bel University." \
            + "\nThe catalogue may be used only if you refer to the original source." \
            + "\nBibTeX entry:" \
            + "\n" + '-'*88 + "\n" \
            + r"@misc{jk23-1," + "\n" \
            + r"      author       = {Karab\'a\v s, J\'an}," + "\n" \
            + r"      title        = {Actions of finite groups on {R}iemann surfaces of higher genera}," + "\n" \
            + r"      year         = {2023}," + "\n" \
            + r"      howpublished = {\url{https://www.savbb.sk/~karabas/science/discactions.html}}" + "\n" \
            + r"}"  \
            + "\n" + '-'*88 + "\n" \
            + r"See also https://www.savbb.sk/~karabas/science/discactions.html for more information." \
            + "\n" + '='*88 + "\n"
        ostrm.write("{}\n\n".format(uvtext))
        ndx = 0
        for elt in action_list:
            ndx += 1
            # if gamma >= 0 and elt['gamma'] != gamma:
            #     continue
            if not(gamma is None):
                if not(elt['gamma'] in gamma):
                    continue 
            if 'top' in elt['enums']:
                if not(multi is None):
                    if not(elt['enums']['top'] in multi):
                        continue
            oline = ""
            # genus, gamma, bd, group, imgs, enums,
            s_label = 'O{}.{}'.format(elt['genus'], ndx)
            oline += s_label
            oline += "\t{}, ".format(f_signature(elt['gamma'], elt['bd']))
            s_gcode, s_grp = f_group(
                elt['group']['order'], elt['group']['code'])
            if s_grp != "":
                oline += "group `{}` of size {}\n".format(
                    s_grp, elt['group']['order'])
            else:
                oline += "small group `{}`\n".format(s_gcode)
            oline += "\t\tEnumerations: {}".format(
                ", ".join(["{}: {}".format(k, v) for k, v in elt['enums'].items()]))
            # andx = 1
            # if 'eqreps' in elt:
            #     for act in elt['eqreps']:
            #         oline += "\t\t{}.{} [{}]\n".format(
            #             s_label, andx, ", ".join(f_permutation(l) for l in act))
            #         andx += 1
            #     ostrm.write("{}\n".format(oline))
            andx = 1
            if 'autgreps' in elt:
                if trivial and len(elt['autgreps']) == 1:
                    oline += ", top: 1\n"
                else:
                    oline += "\n"
                for act in elt['autgreps']:
                    oline += "\t\t{}.{} [{}]\n".format(
                        s_label, andx, ", ".join(f_permutation(l) for l in act))
                    andx += 1
                ostrm.write("{}\n".format(oline))
            else:
                ostrm.write("\n".format(oline))

def parse_range(rng_str = '', positive_only=True):
    if not(rng_str is None) and rng_str != '':
        # m = re.match(r"^(\d+)$", rng_str.strip(),re.IGNORECASE)
        # if m:
        #     # print("Unique: {}", m.groups())
        #     try:
        #         top_g = int(m.groups()[0])
        #         bottom_g = top_g
        #     except:
        #         top_g =  2 ** 10 - 1
        #         bottom_g = 0

        mb = re.match(r"^(\d+)\.\.$", rng_str.strip(),re.IGNORECASE)
        mt = re.match(r"^\.\.(\d+)$", rng_str.strip(),re.IGNORECASE)
        mr = re.match(r"^(\d+)\.\.(\d+)$", rng_str.strip(),re.IGNORECASE)
        ml = re.match(r"^(\d+)(\,(\d+))*$", rng_str.strip(),re.IGNORECASE)



        list_values = False
        if ml:   
            print("List: {}", ml.groups())
            list_values = True
            
            try:
                ret = [int(x) for x in rng_str.split(',') if not(positive_only) or int(x) >= 0]
            except:
                pass


        elif mr:
            print("Top and bottom: {}", mr.groups())
            try:
                bottom_g = int(mr.groups()[0])
            except:
                bottom_g = 0
            try:
                top_g = int(mr.groups()[1])
            except:
                top_g = 2 ** 32 - 1
        elif mt:
            print("Top: {}", mt.groups())
            try:
                top_g = int(mt.groups()[0])
                if positive_only:
                    bottom_g = 0
                else:
                    bottom_g = - (2**32-1)
            except:
                top_g = 2 ** 32 - 1
        elif mb:
            print("Bottom: {}", mb.groups())
            try:
                bottom_g = int(mb.groups()[0])
                top_g = 2**32-1
            except:
                if positive_only:
                    bottom_g = 0
                else:
                    bottom_g = -(2**32-1)
        else:
            if positive_only:
                bottom_g = 0
            else:
                bottom_g = -(2**32-1)
            top_g = 2**32-1


        if not(list_values):  
            if positive_only:
                if bottom_g < 0:
                    bottom_g = 0
                if top_g < 0:
                    top_g = 0

            if bottom_g > top_g:
                bottom_g, top_g = top_g, bottom_g

            ret = range(bottom_g, top_g+1)
    else:
        return range(-2**32-1,2**32-1)

    return ret
    # print("Arggamma = {}".format(arggamma))

if __name__ == '__main__':
    import argparse
    import sys
    import re
    import string
    import random

    # import secrets
    # import string
    # salt = ''.join((secrets.choice(string.ascii_letters + string.digits) for i in range(6)))

    parser = argparse.ArgumentParser(description='Export census data from JSON',
    formatter_class=argparse.ArgumentDefaultsHelpFormatter,
)

    parser.add_argument('-f','--full',help='Create index and full list file',action='store_true',default=False) 
    # parser.add_argument('-m','--magma',help='Create Magma source',action='store_true',default=False)
    parser.add_argument('-i','--index',help='Disable creation of index file',action='store_true',default=False)
    parser.add_argument('-t','--tex',help='Create TeX source',action='store_true',default=False)
    parser.add_argument('-s','--salt',help='Randomize name',action='store_true',default=False)
    parser.add_argument('-g','--gamma', help="Include action of given genera only", type=str)
    parser.add_argument('-m','--multi', help="Include action of given multiplicity only", type=str)
    parser.add_argument('--trivial', help="Set ~top=1 if autg=1", action='store_true',default=False)
    parser.add_argument('path', metavar='PATH', type=str, nargs=1,help='The filename')
    args = parser.parse_args()

    create_index = not(args.index)
    create_tex = args.tex
    create_full = args.full

    # print("Create index: {}".format(create_index))
        
    # arggamma = -1
    # if not(args.gamma is None):
    #     arggamma  = args.gamma
    #     if arggamma < 0:
    #         arggamma = -1


    # bottom_g = 0
    # top_g = 2 ** 10 - 1
    arggamma = None
    # print("DEBUG: arggamma {}".format(args.gamma))
    arggamma = parse_range(args.gamma)

    argmulti = None
    # print("DEBUG: argmulti {}".format(args.multi))
    argmulti = parse_range(args.multi)


    salt = ''
    if args.salt:
        salt =  "_" + "".join(random.sample(string.ascii_letters + string.digits, 6))


    try:
        # f = open("actions-genus{}.json".format(args.path[0]),'r')
        f = open(args.path[0], 'r')
        data = json.load(f)
        genus = data['actions'][0]['genus']
        print("I> Census length = {}".format(len(data['actions'])))

        if create_index:
            print("Creating index")
            write_index( data, "actions-genus{}-index{}.txt".format(genus,salt), gamma=arggamma, multi=argmulti, trivial=args.trivial)
        if create_full:
            print("Writing full list")
            write_human_readable( data, "actions-genus{}-full{}.txt".format(genus,salt), gamma=arggamma, multi=argmulti, trivial=args.trivial)
        if create_tex:
            print("Writing TeX")
            write_index( data, "actions-genus{}{}.tex".format(genus,salt), gamma=arggamma, multi=argmulti, latex=True, trivial=args.trivial)


        f.close()
    except Exception as e:
        print(e)
        traceback.print_exc()
        sys.exit(1)
